"""Generates multivariate normally distributed data.

Module taken from `https://github.com/ralphma1203/trun_mvnt <https://github.com/ralphma1203/trun_mvnt>`_.
"""
import math
import numpy as np

from scipy.stats import norm, weibull_min

##### Functions for acceptance rate #####
# Acceptance rate of normal rejection sampling
def norm_acc(a, b):
    return (norm.cdf(b) - norm.cdf(a))


# Acceptance rate of half-normal rejection sampling
def halfnorm_acc(a, b):
    return 2*norm_acc(a, b)

# Acceptance rate of uniform rejection sampling
def unif_acc(a,b):
    constant = math.sqrt(2*math.pi)/(b-a)
    cp = constant*norm_acc(a, b)
    
    if (a <= 0 and b >= 0):
        return cp
    
    if a > 0:
        return cp*math.exp(a**2/2)
    
# Acceptance rate of translated-exponential rejection sampling
def exp_acc_opt(a,b):
    lam = (a + math.sqrt(a**2 + 4))/2
    return math.sqrt(2*math.pi)*norm_acc(a, b)*lam \
        *math.exp(-lam**2/2 + lam*a)

##### Functions for rejection sampling #####
# Normal rejection sampling
def norm_rej(a, b=math.inf):
    
    acc = 0
    
    while 1:
        x = norm.rvs()
        acc += 1 # one more draw
        
        if x >= a and x <= b:
            d = dict()
            d['x'] = x ; d['acc'] = acc
            return d
    
# Half-normal rejection sampling
def halfnorm_rej(a, b):
    acc = 0
    
    while 1:
        x = norm.rvs()
        acc += 1 # one more draw
        
        if abs(x) >= a and abs(x) <= b:
            d = dict()
            d['x'] = abs(x); d['acc']   = acc
            return d

# Uniform rejection sampling
def unif_rej(a,b):
        
    if 0 >= a and 0 <= b: type_ab = '1'
    if a > 0: type_ab = '2'
    if b < 0: type_ab = '3'
    
    # different cases for the ratio
    def rho_fun(type_ab):
        return {
            '1': math.exp(-x**2/2),
            '2': math.exp(-(x**2 - a**2)/2),
            '3': math.exp(-(x**2 - b**2)/2)
            }[type_ab]
        
    acc = 0
    
    while 1:
        x = np.random.uniform(low=a, high=b, size=1)
        u = np.random.uniform()
        
        acc += 1
        
        rho = rho_fun(type_ab)
        if u <= rho:
            d = dict()
            d['x'] = x; d['acc'] = acc
            return d
        
# Translated-exponential rejection sampling
def exp_rej(a, b=math.inf, lam_type='default'):
    if lam_type == 'default':
        lam = a
    else:
        lam = (a + math.sqrt(a**2 + 4))/2
    
    # rho is continuous at lambda=a, 
    # so don't need to change the expression of rho
    
    acc = 0
    
    while 1:
        x = weibull_min.rvs(1, loc=0, scale=1/lam) + a
        u = np.random.uniform()
        rho = math.exp(-(x-lam)**2/2)
        
        acc += 1
        
        if u <= rho and x < b:
            d = dict()
            d['x'] = x; d['acc'] = acc
            return d


##### Functions for pdf/cdf of truncated univariate distribution #####

# Density function of truncated univariate normal distribution
def dtuvn(x, mean, sd, lower, upper):
    if x < lower or x > upper:
        return 0
    else:
        value = norm.pdf(x, mean, sd)
        value = value/(norm.cdf(upper, mean, sd) - norm.cdf(lower, mean, sd))
        return value

# Distribution function of truncated univariate normal distribution
def ptuvn(x, mean, sd, lower, upper):
    if x <= lower: return 0
    if x >= upper: return 1
    
    num = norm.cdf(x, mean, sd) - norm.cdf(lower, mean, sd)
    dem = norm.cdf(upper, mean, sd) - norm.cdf(lower, mean, sd)
    
    return num/dem


##### Rejection sampling of truncated univariate normal #####

# Rejection sampling of standardized truncated univariate normal distribution
def imp(a,b):
    
    def lowerb1(a):
        return (math.sqrt(math.pi/2)*math.exp(a**2/2) + a)
    
    # lower bound of b for normal vs unif in 0 in [a,b]
    def lowerb2(a):
        return (math.sqrt(2*math.pi) + a)
    
    # lower bound of b for exp vs unif in [a,b]>=0
    def lowerb(a):
        lam = (a + math.sqrt(a**2 + 4))/2
        value = a + math.exp(0.5)/lam \
            *math.exp(( a**2 - a*math.sqrt(a**2 + 4) )/4 )
        
        return value
    
    # Case 1: [a,infty)
    def imp_case1(a,b=math.inf):
        if a < 0:
            return norm_rej(a,b) 
        else:
            if a < 0.25696:
                return halfnorm_rej(a,b)
            else:
                return exp_rej(a,b, 'opt')
            
    # Case 2: 0 in [a,b], a<0<b
    def imp_case2(a,b):
        if b > lowerb2(a):
            return norm_rej(a,b)
        else:
            return unif_rej(a,b)
            
        
    # Case 3: [a,b], a>=0
    def imp_case3(a,b):
        if a <= 0.25696:
            blower1 = lowerb1(a)
            if b <= blower1:
                return unif_rej(a,b)
            else:
                return halfnorm_rej(a,b)
            
        else:
            blower2 = lowerb(a)
            if b <= blower2:
                return unif_rej(a,b)
            else:
                return exp_rej(a,b,'opt')
            
    # Case 4: (-infty,b] (symmetric to Case 1)
    def imp_case4(a,b):
        temp = imp_case1(-b, -a)
        x = -temp['x']
        
        d = dict()
        d['x'] = x; d['acc'] = temp['acc']
        return d
    
    # Case 5: [a,b], b<=0 (symmetric to Case 3)
    def imp_case5(a,b):
        temp = imp_case3(-b,-a)
        x = -temp['x']
        
        d = dict()
        d['x'] = x; d['acc'] = temp['acc']
        return d
    

    if a == -math.inf or b == math.inf:
        if b == math.inf:
            return imp_case1(a,b)
        else:
            return imp_case4(a, b)
    else:
        if a < 0 and b > 0 : return imp_case2(a,b)
        if a >= 0: return imp_case3(a,b)
        if b <= 0: return imp_case5(a,b)
        

# Acceptance rate of truncated univariate normal distribution rejection sampling
def imp_acc(a,b):
    def lowerb1(a):
        return math.sqrt(math.pi/2)*math.exp(a**2/2) + a
    
    # lower bound of b for normal vs unif in 0 in [a,b]
    def lowerb2(a):
        return math.sqrt(2*math.pi) + a
    
    # lower bound of b for exp vs unif in [a,b]>=0
    def lowerb(a):
        lam = (a + math.sqrt(a**2 + 4))/2
        value = a + math.exp(0.5)/lam \
            *math.exp(( a**2 - a*math.sqrt(a**2 + 4) )/4 )
        
        return value
  
    # Case 1: [a,infty)
    def imp_acc_case1(a,b):
        if a <0: 
            return norm_acc(a,b)
        else:
            if a < 0.25696:
                return halfnorm_acc(a,b)
            else:
                return exp_acc_opt(a,b)

    # Case 2: 0 in [a,b], a<0<b
    def imp_acc_case2(a,b):
        if b > lowerb2(a):
            return norm_acc(a,b)
        else:
            return unif_acc(a,b)
    
    # Case 3: [a,b], a>=0
    def imp_acc_case3(a,b):
        if a <= 0.25696:
            blower1 = lowerb1(a)
            if b <= blower1:
                return unif_acc(a,b)
            else:
                return halfnorm_acc(a,b)
        else:
            blower2 = lowerb(a)
            if b <= blower2:
                return unif_acc(a,b)
            else:
                return exp_acc_opt(a,b)
            
    # Case 4 and 5 are symmetric to 1 and 3
    
    if a == -math.inf or b == math.inf:
        if b == math.inf:
            return imp_acc_case1(a)
        else:
            return imp_acc_case1(-b)
    else:
        if a < 0 and b > 0: return imp_acc_case2(a,b)
        if a >= 0: return imp_acc_case3(a,b)
        if b <= 0: return imp_acc_case3(-b, -a)
        

# Random number generation for truncated univariate normal distribution
def rtuvn(n=1, mean=0, sd=1, lower=-math.inf, upper=math.inf):
    
    
    # transform the boundaries
    a = (lower - mean)/sd
    b = (upper - mean)/sd

    # generate n samples from TN(0,1;a,b)
    Z = np.empty(n)
    for i in range(n):
        temp = imp(a,b)
        Z[i-1] = temp['x']
        
    return sd*Z + mean


##### Core functions: Random number generation for truncated multivariate normal/t distribution #####

#' Random number generation for truncated multivariate normal distribution subject to linear inequality constraints
def rtmvn(n=1, Mean=np.zeros(1), Sigma=np.ones(1), \
          D=np.ones(1), \
          lower=np.array(-math.inf), upper=np.array(math.inf), \
          burn=10, thin=1, ini=[]):
    
    D = np.matrix(D); [m,p] = D.shape # m constraints p-dimensional TN
    Sigma = np.matrix(Sigma)
    
    Mean = np.matrix(Mean)
    lower = np.matrix(lower); upper = np.matrix(upper)
    
    
    if Sigma.shape[1] != p or Sigma.shape[0] != Sigma.shape[1]: # check D and Sigma dimension
        print('No. col of "D" should match with dim. of "Sigma", and "Sigma" should be square matrix')
        return []
    
    if Mean.size != p or lower.size != m or upper.size != m : # check Mean an D matrix dimension 
        print('dimension of "Mean should match with No. col of "D". Dimension of "lower" and "upper" should match with No. row of "D"!')
        
        return []
    else:
        Mean = Mean.reshape((p,1))
        lower = lower.reshape((m,1)); upper = upper.reshape((m,1))
    
    
    if Mean.shape == (1,1):
        if lower >= upper:
            print('"lower" should be smaller than "upper"!')
            return []
        else:
            return np.array(rtuvn(n, Mean, math.sqrt(Sigma[0]), lower, upper))
    else:
        
      
        if any(lower >= upper) : # check any lower >= upper
            print('lower bound must be smaller than upper bound!')
            return []
        
        bound_check = 0
        
        if np.array(ini).size == p :
            ini_test = np.matmul(D, ini)
            lower_log = ini_test >= (lower + 1e-8)
            upper_log = ini_test <= (upper - 1e-8)
            bound_check = np.prod(np.multiply(lower_log,upper_log))
               
            if bound_check == 0:
                print('initial is outside or too close from boundary, will be auto-corrected by MP-inverse!')
        elif bound_check == 0:
            ini = np.matmul(np.linalg.pinv(D), (lower + upper)/2)
            
    if any(np.array([burn,thin,n]) % 1 != 0) :
        print('burn, thin and n must be integer!\n')
        return []
    if any(np.array([burn,thin,n-1]) % 1 != 0):
        print("burn, thin must be non-negative interger, n must be positive integer\n")
    
    a = lower - np.matmul(D, Mean)
    b = upper - np.matmul(D, Mean)
    Sigma_chol = np.linalg.cholesky(Sigma)
    R = np.matmul(D, Sigma_chol)
    
    
    z = np.linalg.solve(Sigma_chol, ini - Mean)
    
    keep_x = np.zeros( ((thin+1)*n+burn, p) )

    
    for i in range( (thin+1)*n+burn ):
        for j in range(p):
            rj = R[:,j]           # the jth column of R
            Rj = np.delete(R,j,1) # m by p-1 matrix by removing the jth column of R
            zj = np.delete(z,j).reshape((p-1,1))  # p-1 vector by removing the jth element
            a_temp = a - np.matmul(Rj, zj)
            b_temp = b - np.matmul(Rj, zj)
            
            # ignoring rj = 0, as alwyays fulfill
            pos = np.array(rj > 0)
            neg = np.array(rj < 0)
            
            if pos.sum() == 0:
                lower_pos = -math.inf ; upper_pos = math.inf
            else:
                lower_pos = (a_temp[pos]/rj[pos]).max() # when r_jk>0, a associated with lower bound
                upper_pos = (b_temp[pos]/rj[pos]).min() #              b                 upper
           
            if neg.sum() == 0:
                upper_neg = math.inf; lower_neg = -math.inf
            else:
                upper_neg = (a_temp[neg]/rj[neg]).min() # when r_jk<0, a                 upper
                lower_neg = (b_temp[neg]/rj[neg]).max() #              b                 lower
            
            lower_j = max(lower_pos, lower_neg)
            upper_j = min(upper_pos, upper_neg)
            
            z[j] = rtuvn(n=1, mean=0, sd=1, lower=lower_j, upper=upper_j)
            
        keep_x[i,:] = (np.matmul(Sigma_chol,z) + Mean).reshape(1,p)
    
    final_ind = np.array(range((thin+1)*n))
    final_ind = final_ind[0:len(final_ind):(thin+1)] + thin + burn
    
    return keep_x[final_ind,:]


#' Random number generation for truncated multivariate Student's t distribution subject to linear inequality constraints
def rtmvt(n=1, Mean=np.zeros(1), Sigma=np.ones(1), nu=100, \
          D=np.ones(1), \
          lower=np.array(-math.inf), upper=np.array(math.inf), \
          burn=10, thin=1, ini=[]):
    D = np.matrix(D); [m,p] = D.shape # m constraints p-dimensional TN
    Sigma = np.matrix(Sigma)
    
    Mean = np.matrix(Mean)
    lower = np.matrix(lower); upper = np.matrix(upper)
    
    
    if Sigma.shape[1] != p or Sigma.shape[0] != Sigma.shape[1]: # check D and Sigma dimension
        print('No. col of "D" should match with dim. of "Sigma", and "Sigma" should be square matrix')
        return []
    
    if Mean.size != p or lower.size != m or upper.size != m : # check Mean an D matrix dimension 
        print('dimension of "Mean should match with No. col of "D". Dimension of "lower" and "upper" should match with No. row of "D"!')
        
        return []
    else:
        Mean = Mean.reshape((p,1))
        lower = lower.reshape((m,1)); upper = upper.reshape((m,1))
    
    
    if Mean.shape == (1,1):
        if lower >= upper:
            print('"lower" should be smaller than "upper"!')
            return []
        else:
            return np.array(rtuvn(n, Mean, math.sqrt(Sigma[0]), lower, upper))
    else:
        
      
        if any(lower >= upper) : # check any lower >= upper
            print('lower bound must be smaller than upper bound!')
            return []
        
        bound_check = 0
        
        if np.array(ini).size == p :
            ini_test = np.matmul(D, ini)
            lower_log = ini_test >= (lower + 1e-8)
            upper_log = ini_test <= (upper - 1e-8)
            bound_check = np.prod(np.multiply(lower_log,upper_log))
            
            if bound_check == 0:
                print('initial is outside or too close from boundary, will be auto-corrected by MP-inverse!')
        elif bound_check == 0:
            ini = np.matmul(np.linalg.pinv(D), (lower + upper)/2)
            
    if any(np.array([burn,thin,n]) % 1 != 0) :
        print('burn, thin and n must be integer!\n')
        return []
    if any(np.array([burn,thin,n-1]) % 1 != 0):
        print("burn, thin must be non-negative interger, n must be positive integer\n")
    
    a = lower - np.matmul(D, Mean)
    b = upper - np.matmul(D, Mean)
    Sigma_chol = np.linalg.cholesky(Sigma)
    R = np.matmul(D, Sigma_chol)
    
    x = np.linalg.solve(Sigma_chol, ini - Mean) # initial value for the transformed tmvt
    
    keep_t = np.zeros( ((thin+1)*n+burn, p) )
    
    for i in range((thin+1)*n+burn):
        u = np.random.chisquare(df=nu)
        denom = math.sqrt(u/nu)
        
        lw = a*denom ; up = b*denom
        z0 = x*denom
        
        z = np.matrix(rtmvn(n=1, Mean=np.zeros(p), Sigma=np.diag(np.ones(p)), \
                  D=R, lower=lw, upper=up, burn=0, ini=z0)).reshape((p,1))
        
        keep_t[i,:] = (np.matmul(Sigma_chol,z/denom) + Mean).reshape(1,p)

    final_ind = np.array(range((thin+1)*n))
    final_ind = final_ind[0:len(final_ind):(thin+1)] + thin + burn
    
    return keep_t[final_ind,:]    

 


"""
Computes generalization error of optimization risk minimizer.

In the excess risk decomposition setup, a constrained set of functions must be
considered to reduce the incidence of overfitting. Because sample sizes are limited
in practice, the best possible function in the constrained set of functions (i.e.,
constrained empirical risk minimizer, :math:`f_\\mathcal{F}`) is unlikely to be obtained. Instead, an estimate
of the constrained empirical risk minimizer will result and will have a larger
generalization error than the constrained empirical risk minimizer. I call this estimate
the sample risk minimizer, :math:`\\hat{f}_s`. If no closed-form solution exists for
computing parameter estimates, then only an estimate of the sample risk minimizer will be
obtained, which I call the optimization risk minimizer, :math:`\\tilde{f}_s`.
To the extent that optimization is imperfect, the generalization error of the optimization
risk minimizer will be different from the sample risk minimizer (note that
the optimization risk minimizer can have a generalization error that is larger or smaller
than the sample risk minimizer. This module computes the generalization error of the optimization risk
minimizer. For more details, see whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

References
----------
[1] Bottou, L. & Bousquet, O. (2007). The tradeoffs of large scale learning. InKoller, S.
(Eds.), Advances in neural information processing systems.
(pp. 161–168). Curran Associates, Inc. `bit.ly/3qo1xpI <bit.ly/3qo1xpI>`_.
"""

import numpy as np
from smltheory.overfitting import compute_all_emp_gen_errors
from sklearn.metrics import mean_squared_error
from smltheory.gradient_descent import gradient_descent
from smltheory.least_squares import extract_feature_outcome_data


def determine_best_polynomial_model(data_sample, data_gen_error,
                                    poly_order_range=range(1, 6)):
    """Determines polynomial model with lowest generalization error.

    Parameters
    --------
    data_sample: pandas.core.frame.DataFrame
       Sample size to use for empirical data (can be conceptualized as either the training or
       validation set).

    data_gen_error: pandas.core.frame.DataFrame
        Data set used to estimate generalization of sample risk minimizer. Similar to the data
        set specified for the data_best_in_class argument, the data set here should likewise be
        large so that accurate estimate is obtained for a function's generalization error.

    poly_order_range: range
        Range of polynomial functions over which to conduct empirical risk minimization.

    Returns
    --------
    best_poly_order: numpy.int64
        Polynomial order of best model.

    See Also
    --------
    generate_data.generate_mult_trunc_normal()

    Examples
    --------
    >>> from smltheory.generate_data import create_covariance_matrix, generate_mult_trunc_normal
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #specify sample sizes for data sets
    >>> sample_size_gen_error = 150
    ... sample_size_data_best_in_class = 1e4
    #generate data sets
    >>> data_best_in_class = generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
    ... sample_size=sample_size_data_best_in_class, seed=7)
    >>> data_gen_error = generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
    ... mu=mu, sample_size=sample_size_gen_error, seed = 22
    >>> data_sample = data_best_in_class.sample(n=50, random_state=27)
    #determine best polynomial model
    >>> best_model_poly_order = determine_best_polynomial_model(data_sample=data_sample,
    ... data_gen_error=data_gen_error)
    """
    # Use random_state to ensure reproducibility and prevent resampling from adding noise to estimates
    df_all_emp_gen_errors = compute_all_emp_gen_errors(data_emp_loss=data_sample,
                                                       data_gen_error=data_gen_error,
                                                       include_interactions=False,
                                                       poly_order_range=poly_order_range)

    # Identify best polynomial model
    best_poly_order = df_all_emp_gen_errors["poly_order"][df_all_emp_gen_errors["gen_error"].idxmin()]

    return best_poly_order


def compute_opt_gen_error(opt_weights, poly_order, data_gen_error):
    """Computes generalization error of best polynomial model.

    Parameters
    --------
    opt_weights: numpy.ndarray
        Regression weights of polynomial model.

    poly_order: int
        Polynomial order of model.

    data_gen_error: pandas.core.frame.DataFrame
        Data set used to estimate generalization of sample risk minimizer. Similar to the data
        set specified for the data_best_in_class argument, the data set here should likewise be
        large so that accurate estimate is obtained for a function's generalization error.

    Returns
    --------
    gen_error: numpy.int64
        Generalization error of regression model with weights defined by `opt_weights`.

    See Also
    --------
    generate_data.generate_mult_trunc_normal()

    Examples
    --------
    >>> from smltheory.generate_data import *
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #specify sample sizes for data sets
    >>> sample_size_gen_error = 150
    ... sample_size_data_best_in_class = 1e4
    #generate data sets
    >>> data_best_in_class = generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
    ... sample_size=best_in_class_sample_size, seed=7)
    >>> data_gen_error = generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
    ... mu=mu, sample_size=sample_size_gen_error, seed = 22
    >>> data_sample = data_best_in_class.sample(n=sample_size, random_state=27)
    #determine best polynomial model
    >>> best_model_poly_order = determine_best_polynomial_model(data_sample=data_sample,
    ... data_gen_error=data_gen_error)
    #set initial guess for regression weights
    >>> opt_weights = np.random.uniform(low=0, high=1, size=2 * best_model_poly_order)
    #compute generalization error of optimization risk minimizer
    >>> compute_opt_gen_error(opt_weights=opt_weights, poly_order=poly_order,
    ... data_gen_error=data_gen_error)
    """
    # gather necessary components for matrix-matrix multiplications
    dict_data = extract_feature_outcome_data(data=data_gen_error, poly_order=poly_order)
    features = dict_data["features"]
    outcome = dict_data["outcome"].ravel()

    # compute predictions and generalization error
    predictions = np.dot(features, opt_weights)
    gen_error = mean_squared_error(y_true=outcome, y_pred=predictions)

    return gen_error


def get_opt_risk_min(sample_size, data_best_in_class, data_gen_error, num_iterations=500):
    """Computes generalization error of optimization risk minimizer.

    To obtain generalization error of optimization risk minimizer, two steps are followed:

    1. Determine which polynomial model has the best generalization error.
    2. Compute regression weights for the best polynomial model using a numerical optimization
       method of using gradient descent (instead of a closed-form solution).


    Because this function was constructed to simulate optimization error, gradient descent was
    handicapped so that a set of subtoptimal weights could be obtained with a generalization error
    that differed from the sample risk minimizer, :math:`\\hat{f}_s`.

    Parameters
    --------
    sample_size: numpy.ndarray
        Regression weights of polynomial model.

    data_best_in_class: pandas.core.frame.DataFrame
        Data set used to obtain constrained empirical risk minimizer (i.e., best possible function
        in constrained set of functions). Data set should be large (i.e., >1e4) so that an
        accurate estimate of the contrained empirical risk minimizer is obtained.

    data_gen_error: pandas.core.frame.DataFrame
        Data set used to estimate generalization of sample risk minimizer. Similar to the data
        set specified for the data_best_in_class argument, the data set here should likewise be
        large so that accurate estimate is obtained for a function's generalization error.

    num_iterations: int
        Number of iterations to use for gradient descent.

    Returns
    --------
    gen_error: numpy.int64
        Generalization error of regression model with weights defined by `opt_weights`.

    See Also
    --------
    generate_data.generate_mult_trunc_normal()

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #specify sample sizes for data sets
    >>> sample_size_gen_error = 150
    ... sample_size_data_best_in_class = 1e4
    #generate data sets
    >>> data_best_in_class = generate_data.generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
    ... sample_size=best_in_class_sample_size, seed=7)
    >>> data_gen_error = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
    ... mu=mu, sample_size=sample_size_gen_error, seed = 22
    #compute generalization error of optimization risk minimizer
    >>> get_opt_risk_min(sample_size = 50, data_best_in_class=data_best_in_class, data_gen_error=data_gen_error,
    ... num_iterations=500)
    """
    # Step 1: Using empirical risk minimization to determine the polynomial model order that results in the lowest
    # empirical loss
    data_sample = data_best_in_class.sample(n=sample_size, random_state=27)
    best_model_poly_order = determine_best_polynomial_model(data_sample=data_sample,
                                                            data_gen_error=data_gen_error)

    # Step 2: Obtain regression weights for the polynomial order model by running gradient descent.
    # I fix learning rate to 1e-5 and number of iterations to 500
    np.random.seed(27)
    initial_weights = np.random.uniform(low=0, high=1, size=2 * best_model_poly_order)

    opt_weights = gradient_descent(data=data_sample, initial_weights=initial_weights,
                                   num_iterations=num_iterations, return_log=False, learning_rate=5e-3,
                                   poly_order=best_model_poly_order, expedite_algorithm=True)

    # Step 3: Use the number of columns in features data set to generate random guesses for regression
    # weight starting values
    opt_gen_error = compute_opt_gen_error(opt_weights=opt_weights, poly_order=best_model_poly_order,
                                          data_gen_error=data_gen_error)

    return opt_gen_error



"""Generates data according to second-order polynomial function (and an interaction term).

Data are generated in the context of a wine tasting example. In the example,
weather (:math:`\\mathbf{x}_2`) and winemaking quality (:math:`\\mathbf{x}_2`) are treated as
predictors of wine quality (:math:`\\mathbf{y}`). Note that, together, weather and winemaking
quality make up the feature space (:math:`\\mathbf{X}`). Specifically, wine quality is the outcome
of the following second-order polynomial model:

:math:`\\begin{align}
\\mathbf{y} &= \\mathbf{Xw} + \\mathbf{e}, \\text { where} \\\\
\\mathbf{X} &= [\\mathbf{x}_1, \\mathbf{x}_2, \\mathbf{x}_1^2, \\mathbf{x}_2^2, \\mathbf{x}_1\\mathbf{x}_2], \\\\
\\mathbf{w} &= [0.3 ,  0.1 ,  0.07, -0.1 ,  0.1], \\text{ and} \\\\
\\mathbf{e} &\\sim \\mathcal{N}(\\mu = 0, \\sigma = 1.5).
\\end{align}`

For more details, see whitepaper at
`sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
"""

import numpy as np
import pandas as pd
from . import trun_mvnt as tvm


def provide_weights():
    """Returns weights of Bayes decision function.

    Returns
    --------
    weights: numpy.ndarray
        Weights of Bayes decision function.

    Examples
    --------
    >>> provide_weights()
    """

    b_weight_weather = 0.3
    b_weight_winemaking = 0.1
    b_weight_weather_squared = 0.07
    b_weight_winemaking_squared = -0.1
    b_weight_weather_winemaking = 0.1

    weights = np.array([b_weight_weather, b_weight_winemaking,
                        b_weight_weather_squared, b_weight_winemaking_squared, b_weight_weather_winemaking])

    return weights


def compute_old_scale_range():
    """Computes original scale of outcome variable.

    Data for features and outcome were generate to be on a 1--10 scale. To convert the outcome variable of
    wine quality to a 1--10 scale, the original or old range must be obtained. To do so, I calculated
    the lower and upper limits and computed the difference. For more details,see
    `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.


    Returns
    --------
    old_upper_limit - old_lower_limit: numpy.float.64
        Old range of outcome variable.

    Examples
    --------
    >>> compute_old_scale_range()
    """

    # assume highest/lowest possible error scores are +-3 SDs
    old_lower_limit = np.dot(a=provide_weights(), b=np.array([1, 10, 1, 100, 10])) - 3 * 1.5
    old_upper_limit = np.dot(a=provide_weights(), b=np.array([10, 10, 100, 100, 100])) + 3 * 1.5

    return old_upper_limit - old_lower_limit


def compute_rescaling_factor(new_upper_limit=10, new_lower_limit=1):
    """Computes rescaling factor for converting data from old to new scale.

    Rescaling factor is computed from scale of original predictors of weather and winemaking quality
    to new scale with desired lower and upper limits. Rescaling factor is computed according to following

    :math:`\\begin{align}
    \\text{Scaling factor} &= \\frac{\\text{New scale range}}{\\text{Old scale range}}
    \\end{align}`

    For more information on how the old scale range is computed, see whitepaper at
    `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

    Parameters
    --------
    new_upper_limit: int
        Upper limit of new scale.
    new_lower_limit: int
        New lower limit

    Returns
    --------
    rescaling_factor: numpy.float64
        Factor for rescaling from old scale to new scale.

    Examples
    --------
    >>> compute_rescaling_factor(new_upper_limit=5, new_lower_limit=10)
    """
    old_scale_range = compute_old_scale_range()
    new_scale_range = new_upper_limit - new_lower_limit

    rescaling_factor = new_scale_range / old_scale_range

    return rescaling_factor


def create_covariance_matrix(sd, rho):
    """Creates 2x2 covariance matrix.

    Given a vector of two standard deviation values (one for weather and one for winemaking quality),
    :math:`\\mathbf{\\sigma} = [\\mathbf{\\sigma}_1, \\mathbf{\\sigma}_2]`, and a population
    correlation value between the two variables, :math:`\\rho_{\\mathbf{x}_1, \\mathbf{x}_2}`,
    a covariance matrix is computed such that

    :math:`\\begin{align}
    \\boldsymbol{\\Sigma} &= \\begin{bmatrix}
    \\boldsymbol{\\sigma}^2_1  & \\rho_{\\mathbf{x}_1, \\mathbf{x}_2} \\boldsymbol{\\sigma}_1\\boldsymbol{\\sigma}_2 \\\\
    \\rho_{\\mathbf{x}_1, \\mathbf{x}_2} \\boldsymbol{\\sigma}_1\\boldsymbol{\\sigma}_2 & \\boldsymbol{\\sigma}^2_2
    \\end{bmatrix}
    \\end{align}`

    For more information on how the old scale range is computed, see whitepaper at
    `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.


    Parameters
    --------
    sd: list
        List containing two standard deviation values.
    rho: float
        Specifies population-level correlation between two predictor variables of weather and
        winemaking quality.

    Returns
    --------
    cov_matrix: numpy.ndarray
        Covariance matrix for predictors of weather and winemaking quality.

    Examples
    --------
    >>> mu = [5, 7]
    >>> sd = [1.2, 1.7]
    #population correlation between weather and winemaking quality
    >>> rho_weather_winemaking =  0.35
    >>> cov_matrix = create_covariance_matrix(sd = sd, rho =  rho_weather_winemaking)
    """

    # Create a lower triangular matrix with zeros
    n = len(sd)
    cov_matrix = np.zeros((n, n))

    # Fill lower and upper triangles of covariance matrix
    cov_matrix[np.tril_indices(n, -1)] = rho * np.prod(sd)
    cov_matrix = cov_matrix + cov_matrix.T

    # Fill diagonal of covariance matrix
    np.fill_diagonal(a=cov_matrix, val=sd)

    return cov_matrix


def generate_trunc_predictors(mu, cov_matrix, sample_size,
                              lower_limits=1, upper_limits=10, seed=27):
    """Generates predictor variables for weather and winemaking quality.

    Outcome variable of wine quality is rescaled by dividing the desired (or new) scale range by the
    original (or old) scale range. For more information, see whitepaper
    at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

    Parameters
    ---------
    mu: pandas.core.frame.DataFrame
        Data set containing predictors (weather, winemaking quality) and outcome variable (wine quality).
    cov_matrix: int
        Upper limit of new scale.
    sample_size: int
        New lower limit.
    lower_limits: int
    upper_limits: int
    seed: int

    Returns
    --------
    truncated_data:  pandas.core.frame.DataFrame
        Factor for rescaling from old scale to new scale.

    Examples
    --------
    >>> compute_rescaling_factor(new_upper_limit=10, new_lower_limit=2)
    """
    # Upper and lower limits for variables
    lower_limits = np.repeat(lower_limits, len(mu))
    upper_limits = np.repeat(upper_limits, len(mu))

    # Generate samples from multivariate distribution
    sample_size = int(sample_size)
    D = np.diag(np.ones(len(mu)))
    np.random.seed(seed)

    truncated_data = pd.DataFrame(tvm.rtmvn(n=sample_size, Mean=mu, Sigma=cov_matrix,
                                            lower=lower_limits, upper=upper_limits, D=D),
                                  columns=["weather", "winemaking_quality"])

    return truncated_data


def compute_outcome_variable(data):
    """Computes outcome variable (i.e., wine quality).

    Data for outcome variable of wine quality is generated such that

    :math:`\\begin{align}
    \\mathbf{y} &= \\mathbf{Xw} + \\mathbf{e}, \\text { where} \\\\
    \\mathbf{X} &= [\\mathbf{x}_1, \\mathbf{x}_2, \\mathbf{x}_1^2, \\mathbf{x}_2^2, \\mathbf{x}_1\\mathbf{x}_2], \\\\
    \\mathbf{w} &= [0.3 ,  0.1 ,  0.07, -0.1 ,  0.1], \\text{ and} \\\\
    \\mathbf{e} &\\sim \\mathcal{N}(\\mu = 0, \\sigma = 1.5).
    \\end{align}`

    For more details,see `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
        Data set containing predictors (weather, winemaking quality).

    Returns
    --------
    data: pandas.core.frame.DataFrame
        Data set containing predictors (weather, winemaking quality) and outcome variable (wine quality).

    See Also
    --------
    generate_data.create_covariance_matrix()
    generate_data.generate_trunc_predictors()

    Examples
    --------
    >>> mu = [5, 7]
    >>> sd = [1.2, 1.7]
    >>> rho_weather_winemaking =  0.35
    >>> cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    >>> sample_size_gen_error = 150
    >>> sample_size_data_best_in_class = 1e4
    >>> data = generate_trunc_predictors(mu=mu, cov_matrix=cov_matrix,
    ... sample_size=sample_size, seed=27)
    >>> compute_outcome_variable(data=data)
    """

    feature_cols = pd.concat(objs=[data, data ** 2, data.prod(axis=1)], axis=1)

    # Error
    error = np.random.normal(loc=0, scale=1.5, size=data.shape[0])

    # Compute outcome variable of wine quality
    data["wine_quality"] = np.dot(a=feature_cols, b=provide_weights()) + error

    return data


def rescale_outcome_variable(data, new_lower_limit=1, new_upper_limit=10):
    """Rescales outcome variable to have scale defined by new lower and upper limits.

    Outcome variable of wine quality is rescaled by dividing the desired (or new) scale range by the
    original (or old) scale range. For more information, see whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

    Parameters
    --------
    data: pandas.core.frame.DataFrame
        Data set containing predictors (weather, winemaking quality) and outcome variable (wine quality).
    new_upper_limit: int
        Upper limit of new scale.
    new_lower_limit: int
        New lower limit.

    Returns
    --------
    data: pandas.core.frame.DataFrame
        Data set containing predictors (weather, winemaking quality) and rescaled outcome variable (wine quality).

    Examples
    --------
    >>> compute_rescaling_factor(new_upper_limit=10, new_lower_limit=2)
    """
    rescaling_factor = compute_rescaling_factor(new_upper_limit=new_upper_limit,
                                                new_lower_limit=new_lower_limit)
    old_lower_limit = np.dot(a=provide_weights(), b=np.array([1, 10, 1, 100, 10])) - 3 * 1.5
    lower_limit_center = (data["wine_quality"] - old_lower_limit)

    data["wine_quality"] = pd.DataFrame(data=new_lower_limit + lower_limit_center * rescaling_factor,
                                        columns=["wine_quality"])

    return data


def generate_mult_trunc_normal(cov_matrix, mu, sample_size, seed=27):
    """Generates multivariate normal truncated data.

    Data for features (weather, winemaking quality) and outcome (wine quality) are generated according
    to truncated normal distrib utions. For more details,see `
    sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

    Parameters
    ----------
    cov_matrix: pandas.core.frame.DataFrame
        Data set containing predictors (weather, winemaking quality).
    mu: list
        Specifies population-level mean for features.
    sample_size: int
        Sample sizeof generated data set.
    seed:
        Seed value of random number generator.

    Returns
    --------
    data_mult_trunc_normal: pandas.core.frame.DataFrame
        Data set containing predictors (weather, winemaking quality) and outcome variable (wine quality),
        with data for the predictors and outcome rescaled on the desired range.

    See Also
    --------
    generate_data.generate_trunc_predictors()
    generate_data.compute_outcome_variable()
    generate_data.rescale_outcome_variable
    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #specify sample sizes for data sets
    >>> generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    """
    # generate predictors
    data_mult_trunc_normal = generate_trunc_predictors(mu=mu, cov_matrix=cov_matrix,
                                                       sample_size=sample_size, seed=seed)

    # generate outcome variable
    data_mult_trunc_normal = compute_outcome_variable(data=data_mult_trunc_normal)

    # scale outcome variable
    data_mult_trunc_normal = rescale_outcome_variable(data=data_mult_trunc_normal)

    return data_mult_trunc_normal



"""Computes generalization error of Bayes decision function.

To provide context, Bayes decision function predicts wine quality using weather and
winemaking quality such that

:math:`\\begin{align}
\\mathbf{y} &= \\mathbf{Xw}, \\text { where} \\\\
\\mathbf{X} &= [\\mathbf{x}_1, \\mathbf{x}_2, \\mathbf{x}_1^2, \\mathbf{x}_2^2, \\mathbf{x}_1\\mathbf{x}_2], \\\\
\\mathbf{w} &= [0.3 ,  0.1 ,  0.07, -0.1 ,  0.1].
\\end{align}`

For more details, see whitepaper at
`sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
"""

import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error
from smltheory.generate_data import provide_weights, compute_rescaling_factor


def generate_bayes_features(data):
    """Generates features for Bayes decision function.

     Bayes decision function for wine quality was set to be the outcome a second-order
     polynomial model of weather and winemaking quality and the interaction between the
     first-order version of the predictors.

     Parameters
     -----------
     data : pandas.core.frame.DataFrame
        A dataframe created using `generate_mult_trunc_normal()` from the `generate_data`
        module.

     Returns
     -------
     bayes_features : pandas.core.frame.DataFrame
         A dataframe that contains features needed to compute Bayes decision function predictions

     See Also
     --------
     generate_data.generate_mult_trunc_normal()

     Examples
     ---------
     >>> mu=[5, 7]
     >>> sd=[1.2, 1.7]
     >>> rho_weather_winemaking =  0.35
     >>> cov_matrix=generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
     >>>sample_size_gen_error = 150
     >>> data_gen_error = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                          mu=mu, sample_size=sample_size_gen_error)
     >>> compute_bayes_features(data=data_gen_error)
     """

    # Feature columns
    predictors = data[["weather", "winemaking_quality"]]
    bayes_features = pd.concat(objs=[predictors, predictors**2, predictors.prod(axis=1)], axis=1)

    return bayes_features


def compute_bayes_intercept(data):
    """Computes intercept for Bayes decision function.

    Bayes decision function for wine quality was set to be the outcome a second-order
    polynomial model of weather and winemaking quality and the interaction between the
    first-order version of the predictors.

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
       A pd.DataFrame created using `generate_mult_trunc_normal()` from the `generate_data`
       module

    Returns
    ---------
    intercept: np.array
        Intercept value of Bayes decision function.

    See Also
    ---------
    generate_data.generate_mult_trunc_normal()

    Examples
    ---------
    >>> mu = [5, 7]
    >>> sd = [1.2, 1.7]
    >>> rho_weather_winemaking =  0.35
    >>> cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    >>> sample_size_gen_error = 150
    >>> data_gen_error = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                         mu=mu, sample_size=sample_size_gen_error)
    >>> compute_bayes_intercept(data=data_gen_error)
    """

    bayes_features = generate_bayes_features(data=data)

    # Intercept
    intercept = np.mean(data["wine_quality"]) - np.sum(provide_weights() * compute_rescaling_factor() *
                                                       np.mean(bayes_features, axis=0))

    return intercept


def compute_bayes_risk(data):
    """Computes generalization error/risk of Bayes decision function .

    Bayes decision function for wine quality was set to be the outcome a second-order
    polynomial model of weather and winemaking quality and the interaction between the
    first-order version of the predictors.

    Parameters
    --------
    data: pandas.core.frame.DataFrame
       A pd.DataFrame created using `generate_mult_trunc_normal()` from the `generate_data`
       module

    Returns
    --------
    bayes_risk: numpy.float64
        Generalization error/risk of Bayes decision function.

    See Also
    --------
    generate_data.generate_mult_trunc_normal()

    Examples
    --------
    >>> mu = [5, 7]
    >>> sd = [1.2, 1.7]
    >>> rho_weather_winemaking =  0.35
    >>> cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    >>> sample_size_gen_error = 150
    >>> data_gen_error = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                         mu=mu, sample_size=sample_size_gen_error)
    >>> compute_bayes_risk(data=data_gen_error)
    """

    bayes_features = generate_bayes_features(data=data)

    # Compute intercept
    intercept = compute_bayes_intercept(data=data)

    # Compute predictions for wine quality
    test_pred = intercept + np.dot(a=bayes_features, b=provide_weights() * compute_rescaling_factor())

    # Compute mean squared error
    bayes_risk = mean_squared_error(y_true=data['wine_quality'], y_pred=test_pred)

    return bayes_risk



"""Depicts how overfitting results from using complex polynomial models.

Empirical loss and generalization errors are computed for set of polynomial models in a desired range. As
the polynomial order increases, empirical loss remains relatively unchanged, but generalization increases
exponentially. Thus, although more complex models perform well on training data, their actual performance is
considerably worse.

For more details, see
whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
"""

import statsmodels.formula.api as smf
from sklearn.metrics import mean_squared_error
from functools import partial
import pandas as pd


def gen_poly_reg_eq(poly_order, include_interactions=False):
    """Generates equation for polynomial model of specified order.

    Parameters
    ----------
    poly_order: int
        Order of polynomial model to use when creating the set of features.
    include_interactions: bool
        If True, then all possible interactions are included in model.

    Returns
    -------
    equation: str
        String specifying polynomial model.

    Examples
    --------
    >>> gen_poly_reg_eq(poly_order=2)
    'wine_quality ~ np.power(weather, 1) + np.power(weather, 2) + np.power(winemaking_quality, 1) + np.power(winemaking_quality, 2)'
    """
    # compute polynomial terms for predictors
    weather_pred = ' + '.join(['np.power(weather, {})'.format(ord_value) for ord_value in range(1, poly_order + 1)])
    winemaking_quality_pred = ' + '.join(['np.power(winemaking_quality, {})'.format(ord_value)
                                          for ord_value in range(1, poly_order + 1)])

    # Compute all two-way interactions between weather and winemaking_quality
    if include_interactions and poly_order > 1:

        interaction_terms = ' + '.join(
            ['np.multiply(np.power(weather, {}), np.power(winemaking_quality, {}))'.format(p1, p2)
             for p1 in range(1, poly_order)
             for p2 in range(1, poly_order)])

        predictors = ' + '.join([weather_pred, winemaking_quality_pred, interaction_terms])

    else:
        predictors = ' + '.join([weather_pred, winemaking_quality_pred])

    # create regression equation
    equation = "wine_quality ~ {}".format(predictors)

    return equation

def compute_emp_gen_error(equation, data_emp_loss, data_gen_error):
    """Computes empirical loss and generalization error for a polynomial model.

    Parameters
    ----------
    equation: str
        String specifying polynomial model.

    data_emp_loss: pandas.core.frame.DataFrame
        Data set for computing empirical loss.

    data_gen_error: pandas.core.frame.DataFrame
        Data set for computing generalization error.

    Returns
    -------
    emp_loss_gen_error_dict: dict
        Dictionary containing empirical loss and generalization error.

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data_emp_loss = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #sample sizes
    >>> sample_size_emp_loss = 150
    ... sample_size_gen_error = 1e4
    #generate data set for empirical loss
    >>> data_emp_loss = generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
    ... sample_size = sample_size_emp_loss)
    >>> data_gen_error = generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
    ... sample_size = sample_size_gen_error)
    #equation
    >>> equation = gen_poly_reg_eq(poly_order=2)
    >>> compute_emp_gen_error(equation=equation, data_emp_loss=data_emp_loss,
    ... data_gen_error=data_gen_error)
    {'emp_loss': 0.2420572087052563, 'gen_error': 0.25353431587715874}
    """
    model = smf.ols(data=data_emp_loss, formula=equation).fit()
    y_test = model.predict(data_gen_error)

    emp_loss = mean_squared_error(y_true=data_emp_loss['wine_quality'], y_pred=model.fittedvalues)
    gen_error = mean_squared_error(y_true=data_gen_error['wine_quality'], y_pred=y_test)

    emp_loss_gen_error_dict = {'emp_loss': emp_loss, 'gen_error': gen_error}

    return emp_loss_gen_error_dict


def compute_all_emp_gen_errors(data_emp_loss, data_gen_error, poly_order_range, include_interactions=False):
    """Computes empirical loss and generalization error for a polynomial model.

    Parameters
    ----------
    data_emp_loss: pandas.core.frame.DataFrame
        Data set for computing empirical loss.

    data_gen_error: pandas.core.frame.DataFrame
        Data set for computing generalization error.

    poly_order_range: range
        Specifies range of polynomial models to use.

    include_interactions: bool
        If True, polynomial models contain all the corresponding interactions.

    Returns
    -------
    df_emp_gen_errors: pandas.core.frame.DataFrame
        DataFrame containing the empirical loss and generalization error of each polynomial model.

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data_emp_loss = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #sample sizes
    >>> sample_size_emp_loss = 150
    ... sample_size_gen_error = 1e4
    #generate data set for empirical loss
    >>> data_emp_loss = generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
    ... sample_size = sample_size_emp_loss)
    >>> data_gen_error = generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
    ... sample_size = sample_size_gen_error)
    #compute empirical loss and generalization error for each polynomial model
    >>>  compute_all_emp_gen_errors(data_emp_loss = data_emp_loss,
    ... data_gen_error = data_gen_error, include_interactions = True,
    ... poly_order_range=range(1, 10))
        poly_order  emp_loss     gen_error
    0           1  0.244933      0.254885
    1           2  0.242056      0.253487
    2           3  0.235921      0.263245
    3           4  0.219007      0.339122
    4           5  0.202056      2.038340
    5           6  0.184981    537.737954
    6           7  0.177600   6801.505982
    7           8  0.176475  11862.620947
    8           9  0.176777  14885.917939
    """
    # create all polynomial equations within the desired range
    gen_poly_reg_eq_partial = partial(gen_poly_reg_eq, include_interactions=include_interactions)

    poly_equations = list(map(gen_poly_reg_eq_partial, poly_order_range))

    # create partial version of function with data_emp_loss and data_gen_error fixed
    emp_gen_error_partial = partial(compute_emp_gen_error, data_emp_loss=data_emp_loss,
                                    data_gen_error=data_gen_error)

    # for each polynomial equation, compute empirical loss and generalization error
    all_emp_gen_errors = list(map(emp_gen_error_partial, poly_equations))

    # convert dictionary to dataframe and then compute polynomial orders by using row indexes
    df_emp_gen_errors = pd.DataFrame(all_emp_gen_errors)
    poly_orders = pd.Series([poly_order for poly_order in poly_order_range], name="poly_order")

    # concatenate poly_orders and dataframe to create complete dataframe
    df_emp_gen_errors = pd.concat([poly_orders, df_emp_gen_errors], axis=1)

    return df_emp_gen_errors



"""
Computes generalization error of sample risk minimizer.

In the excess risk decomposition setup, a constrained set of functions must be
considered to reduce the incidence of overfitting. Because sample sizes are limited
in practice, the best possible function in the constrained set of functions (i.e.,
constrained empirical risk minimizer, :math:`f_\\mathcal{F}`) is unlikely to be obtained. Instead, an estimate
of the constrained empirical risk minimizer will result and will have a larger
generalization error than the constrained empirical risk minimizer. I call this estimate
the sample risk minimizer, :math:`\\hat{f}_s`. This module computes the generalization error of the optimization risk
minimizer. For more details, see whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

References
----------
[1] Bottou, L. & Bousquet, O. (2007). The tradeoffs of large scale learning. InKoller, S.
(Eds.), Advances in neural information processing systems.
(pp. 161–168). Curran Associates, Inc. `bit.ly/3qo1xpI <bit.ly/3qo1xpI>`_.
"""
from smltheory.overfitting import compute_all_emp_gen_errors


def compute_sample_risk_gen_error(sample_size, data_best_in_class, data_gen_error,
                                  poly_order_range=range(1, 6)):
    """Computes generalization error of sample risk minimizer.

    Parameters
    --------
    sample_size: int
       Sample size to use for empirical data (can be conceptualized as either the training or
       validation set). Note that data of specified sample size is obtained from data set specified
       in data_best_in_class argument.

    data_best_in_class: pandas.core.frame.DataFrame
        Data set used to obtain constrained empirical risk minimizer (i.e., best possible function
        in constrained set of functions). Data set should be large (i.e., >1e4) so that an
        accurate estimate of the contrained empirical risk minimizer is obtained.

    data_gen_error: pandas.core.frame.DataFrame
        Data set used to estimate generalization of sample risk minimizer. Similar to the data
        set specified for the data_best_in_class argument, the data set here should likewise be
        large so that accurate estimate is obtained for a function's generalization error.

    poly_order_range: range
        Range of polynomial functions over which to conduct empirical risk minimization.

    Returns
    --------
    gen_errors.min(): numpy.float64
        Generalization error/risk of sample risk minimizer.

    See Also
    --------
    generate_data.generate_mult_trunc_normal()

    Examples
    --------
    >>> mu = [5, 7]
    >>> sd = [1.2, 1.7]
    >>> rho_weather_winemaking =  0.35
    >>> cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    >>> sample_size_gen_error = 150
    >>> sample_size_data_best_in_class = 1e4

    >>> data_best_in_class = generate_data.generate_mult_trunc_normal(cov_matrix = cov_matrix, mu = mu,
                             sample_size=best_in_class_sample_size, seed=7)
    >>> data_gen_error = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                         mu=mu, sample_size=sample_size_gen_error, seed = 22)
    >>> compute_sample_risk_gen_error(sample_size=50,
                                     data_best_in_class=data_best_in_class,
                                     data_gen_error=data_gen_error)
"""
    # Use random_state to ensure reproducibility and prevent resampling from adding noise to estimates
    gen_errors = compute_all_emp_gen_errors(data_emp_loss=data_best_in_class.sample(n=sample_size, random_state=27),
                                            data_gen_error=data_gen_error,
                                            include_interactions=False,
                                            poly_order_range=poly_order_range)['gen_error']

    # Return generalization error of sample risk minimizer
    return gen_errors.min()



"""Computes cross-section of mean squared error loss function.

To compute the cross-section of the mean squared error loss function, two steps are followed. First, the
closed-form solution  for the regression weights is computed (i.e., the set of regression weights that
minimizes mean squared error). Second, all the weight value from the closed form solution are used except
the first value, and the first value is varied to compute a set of mean squared error values.

For more details, see
whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
"""
import numpy as np
import pandas as pd


def extract_feature_outcome_data(data, sample_size=10, sample_data=False, poly_order=2):
    """Extracts features and outcome variables from data set.

    The features of weather and winemaking quality are extracted and the outcome variable of wine quality
    is extracted.

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
        Data set containing features (weather, winemaking quality) and outcome (wine quality).

    sample_size: int
        If sample_data=True, then a sample of the specified sample size will be drawn randomly from the data
        set.

    sample_data: bool
        If true, only a sample of the specified sample size will be drawn from the data.

    poly_order: int
        Order of polynomial model to use when creating the set of features.

    Returns
    -------
    dict_features_outcome: dict
        Dictionary containing the features and outcome DataFrames.

    See Also
    --------
    generate_data.create_covariance_matrix()
    generate_data.generate_trunc_predictors()

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #extract features and outcome
    >>> extract_feature_outcome_data(data=data)["features"][1:5]
    array([[ 5.45999607,  4.31319487, 29.81155704, 18.60365003],
           [ 4.81312495,  8.83011447, 23.16617174, 77.97092155],
           [ 5.49988075,  6.34522617, 30.24868824, 40.26189513],
           [ 5.23554863,  8.26579874, 27.41096951, 68.32342882]])
    >>> extract_feature_outcome_data(data=data)["outcome"][1:5]
    array([6.85591702, 4.85722482, 5.17003467, 4.65399294])
    """
    if sample_data:
        data = data.sample(n=sample_size, random_state=27)

    # extract set of features and then expand according to polynomial order that is specified
    original_features = data[["weather", "winemaking_quality"]]
    features = pd.concat(objs=[original_features ** order for order in range(1, poly_order + 1)], axis=1).to_numpy()

    outcome = data[["wine_quality"]].to_numpy().ravel()

    dict_features_outcome = {"features": features, "outcome": outcome}

    return dict_features_outcome


def compute_weight_min_mse(data, poly_order=2):
    """Computes closed-form solution for regression weights (i.e., set of weights that minimize mean squared error).

    The following closed-form solution is used to compute regression weights:

    :math:`\\begin{align}
    \\mathbf{w}_{MSE} &= (\\mathbf{X}^\\top\\mathbf{X})^{-1}\\mathbf{X}^\\top\\mathbf{y}.
    \\end{align}`

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
        Data set containing features (weather, winemaking quality) and outcome (wine quality).
    poly_order: int
        Order of polynomial model to use when creating the set of features.

    Returns
    --------
    w_mse.ravel(): numpy.ndarray
        Array that contains regression weights that correspond to closed-form solution.

    See Also
    --------
    generate_data.create_covariance_matrix()
    generate_data.generate_trunc_predictors()

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #compute closed-form solution regression weights
    >>> compute_weight_min_mse(data=data)
    array([ 0.85298795,  1.03699274, -0.03039457, -0.09719031])
    """
    # gather necessary components for matrix-matrix multiplications
    dict_data = extract_feature_outcome_data(data=data, poly_order=poly_order)
    features = dict_data["features"]
    outcome = dict_data["outcome"]

    # compute weight values that result from using closed-form solution
    w_mse = np.linalg.inv(np.transpose(features).dot(features)).dot(np.transpose(features)).dot(outcome)

    return w_mse.ravel()


def compute_ind_mse(data, w_guess, poly_order=2):
    """Computes individual mean squared error value.

    To compute individual mean squared error value, a set of regression weights is needed that
    corresponds to the specified polynomial order. For the later purpose of computing a cross-section of
    the mean squared error function, I fix all the weighht value except the first one to the corresponding
    closed-form solution values.

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
        Data set containing features (weather, winemaking quality) and outcome (wine quality).
    w_guess: numpy.float64
        Weight value to use for first regression weight in computing mean squared error.
    poly_order: int
        Order of polynomial model to use when creating the set of features.

    Returns
    --------
    mse: numpy.float64
        Mean squared error value.

    See Also
    --------
    generate_data.create_covariance_matrix()
    generate_data.generate_trunc_predictors()

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #compute closed-form solution regression weights
    >>> compute_ind_mse(data=data, w_guess=0.34)
    1095.2514584107207
    """
    # gather necessary components for matrix-matrix multiplications
    dict_data = extract_feature_outcome_data(data=data, poly_order=poly_order)
    features = dict_data["features"]
    outcome = dict_data["outcome"]

    # compute weights that result from using closed-form solution
    w_mse = compute_weight_min_mse(data=data, poly_order=poly_order)

    # extract all the values of the closed-form solution except the first value
    w_initial = w_mse[1:]
    w_initial = np.insert(arr=w_initial, obj=0, values=w_guess)

    # compute mean squared error value (i.e., mse = yy - 2wXy + wXXw)
    yy = np.transpose(outcome).dot(outcome)
    wXy = np.transpose(w_initial).dot(np.transpose(features)).dot(outcome)
    wXXw = np.transpose(w_initial).dot(np.transpose(features)).dot(features).dot(w_initial)

    mse = yy - 2 * wXy + wXXw

    return mse


def compute_all_mse_loss_values(data, w_guess_list):
    """Computes cross-section of mean squared error function.

    To compute a cross-section, all but one weight value is fixed. In this case, the first weight value is allowed
    to vary and is set to the value of `w_guess`. All the other weight values are fixed to the values of the
    closed-form solution.

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
        Data set containing features (weather, winemaking quality) and outcome (wine quality).
    w_guess_list: numpy.ndarray
        Set of weight values over which to compute cross-section of mean squared error function.

    Returns
    --------
    df_mse: numpy.float64
        Mean squared error value.

    See Also
    --------
    generate_data.create_covariance_matrix()
    generate_data.generate_trunc_predictors()

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #compute closed-form solution regression weights
    >>> w_guess_list = np.arange(start = 0, stop = 3, step= 0.01)
    ... compute_all_mse_loss_values(data=data, w_guess_list=w_guess_list)
          w_guess     mse_value
    0       0.00   2953.803317
    1       0.01   2885.934168
    2       0.02   2818.865374
    3       0.03   2752.596934
    4       0.04   2687.128850
    ..       ...           ...
    295     2.95  17639.803853
    296     2.96  17808.039463
    297     2.97  17977.075428
    298     2.98  18146.911748
    299     2.99  18317.548423

    [300 rows x 2 columns]
    """
    # return flattened array
    all_mse_loss_values = np.array([compute_ind_mse(data, guess) for guess in w_guess_list])

    # create dataframe with guesses and corresponding MSE values
    df_mse = pd.DataFrame({"w_guess": w_guess_list,
                           "mse_value": all_mse_loss_values})

    return df_mse



"""Runs gradient descent algorithm.

The gradient descent algorithm in this module works by updating an initial set of guesses, :math:`\\mathbf{w}^{(0)}`,
of regression weights and then returning a final set of weights after a specified number of iterations is
reached, :math:`k = K`, or the sum of squared differences between weight values of subsequent iterations is
below some threshold, :math:`\\lVert \\mathbf{w}^{(k)} - \\mathbf{w}^{(k+1)}  \\rVert^2_2 < \\epsilon`.

Note that the mean squared error loss function shown below is used

:math:`\\begin{align}
\\ell_{MSE} &= \\frac{1}{n} \\lVert \\mathbf{y} - \\mathbf{Xw} \\rVert^2_2,
\\end{align}`

and the gradient of the mean squared error loss function, :math:`\\nabla_w \\ell_{MSE}` is computed with respect
to the weights such that

:math:`\\begin{align}
\\nabla_w \\ell_{MSE} &= \\frac{1}{n} \\bigg(2\\mathbf{X}^\\top(\\underbrace{\\mathbf{Xw} -
\\mathbf{y}}_{\\text{Residual}})\\bigg).
\\end{align}`.

Gradient descent works by updating the initial set of weights by stepping in the negative gradient direction
according a set step size (or learning rate, :math:`\\tau`) such that

:math:`\\begin{align}
\\mathbf{w}^{(k+1)} = \\mathbf{w}^{(k)} - \\tau \\nabla_{\\mathbf{w}}\\ell_{MSE}(\\mathbf{w}^{(k)})
\\end{align}`.

Also note that the gradient is a function of the residual, :math:`(\\mathbf{Xw} - \\mathbf{y})`. For more details, see
whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
"""

import numpy as np
import pandas as pd
from smltheory.least_squares import extract_feature_outcome_data, compute_weight_min_mse


def compute_gradient(features, outcome, weights, num_samples):
    """Computes gradient of mean squared error loss function.

    For more details,see `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

    Parameters
    ----------
    features: pandas.core.frame.DataFrame
        Feature used to predict wine quality (weather and winemaking quality).
    outcome:  pandas.core.frame.DataFrame
        Data containing wine quality values.
    weights: numpy.ndarray
        Set of weight values used to compute predictions
    num_samples: int
        Number of data points in either the feature/outcome data sets.

    Returns
    --------
    gradient: pandas.core.frame.DataFrame
        Gradient of mean squared error loss function with respect to regression weights.

    See Also
    --------
    generate_data.create_covariance_matrix()
    generate_data.generate_trunc_predictors()
    least_squares.extract_feature_outcome_data()

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #extract features and outcome
    >>> features = smltheory.least_squares.extract_feature_outcome_data(data=data)["features"]
    ... outcome = smltheory.least_squares.extract_feature_outcome_data(data=data)["outcome"]
    #compute gradient
    >>> weights = np.random.uniform(low = 0, high = 1, size = 4)
    >>> compute_gradient(features=features, outcome=outcome, weights=weights,
    ... num_samples=len(features))
    array([ 691.91945491,  959.77824399, 3812.79874487, 7289.85051885])
    """
    # Compute predictions (Xw using current weights
    predictions = np.dot(features, weights)

    # Compute the error (difference between predictions and target, Xw - y)
    residuals = outcome.ravel() - predictions

    # Compute the gradient of the cost function with respect to weights (2X(Xw - ))
    gradient = -2 * np.dot(features.T, residuals) / num_samples

    return gradient

def gradient_descent(data, initial_weights, learning_rate=0.01, num_iterations=1000, poly_order=2,
                     epsilon=1e-20, expedite_algorithm=False, return_log=False):
    """Runs gradient descent algorithm.

    For a description, see module description at top of page. For complete details see
    `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
        Data set containing features (weather, winemaking quality) and outcome (wine quality).

    initial_weights:  numpy.ndarray
        Array specifiying initial weights.

    learning_rate: float
        Rate at which parameters are updated in each step (i.e., step size parameter)

    num_iterations: int
        Maximum number of iterations to run in gradient descent algorithm.

    poly_order: int
        Order of polynomial order to use in computing gradient descent.

    epsilon: float
        Threshold value set for the difference in the sum of squared differences between weights  between
        subsequent iterations.

    expedite_algorithm: bool
        If True, the gradient descent algorithm will be expedited by replacing all the weight values except the
        first one with the closed-form solution values.

    return_log: bool
        If True, a pandas.DataFrame will be returned that contains the set of weight obtained after each iteration.

    Returns
    --------
    weights: numpy.ndarray
        Set of weights obtained after running gradient descent algorithm.

    See Also
    --------
    generate_data.create_covariance_matrix()
    generate_data.generate_trunc_predictors()
    least_squares.extract_feature_outcome_data()

    Examples
    --------
    >>> mu = [5, 7]
    ... sd = [1.2, 1.7]
    ... rho_weather_winemaking =  0.35
    #generate covariance matrix
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> data = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=150, seed=27)
    #set weights
    >>> np.random.seed(27) #ensure reproducibility
    ... initial_weights = np.random.uniform(low = 0, high = 1, size = 3)
    #run gradient descent algorithm (with no log returned)
    >>> gradient_descent(data=data, initial_weights=initial_weights,  num_iterations=1100,
    ... learning_rate=0.0001, epsilon=1e-20, expedite_algorithm=True, return_log=False)
    array([ 0.85159995,  1.03699274, -0.03039457, -0.09719031])
    #run gradient descent algorithm (with log returned)
    >>> gradient_descent(data=data, initial_weights=initial_weights,  num_iterations=1100,
    ... learning_rate=0.0001, epsilon=1e-20, expedite_algorithm=True, return_log=True)
    iteration_num        w1        w2        w3        w4
    0                 0  0.000000  0.425721  0.814584  0.735397
    1                 1 -0.059489  1.036993 -0.030395 -0.097190
    2                 2 -0.054620  1.036993 -0.030395 -0.097190
    3                 3 -0.049778  1.036993 -0.030395 -0.097190
    4                 4 -0.044961  1.036993 -0.030395 -0.097190
    ...             ...       ...       ...       ...       ...
    1096           1096  0.850382  1.036993 -0.030395 -0.097190
    1097           1097  0.850396  1.036993 -0.030395 -0.097190
    1098           1098  0.850409  1.036993 -0.030395 -0.097190
    1099           1099  0.850423  1.036993 -0.030395 -0.097190
    1100           1100  0.850437  1.036993 -0.030395 -0.097190
    """
    # gather features and outcome data in different sets for later computations
    dict_data = extract_feature_outcome_data(data=data, poly_order=poly_order)
    features = dict_data["features"]
    outcome = dict_data["outcome"]

    # Initialize weights and variable that stores all weights on each iteration
    all_weights = initial_weights  # variable necessary for logging weight values after each iteration
    weights = np.array(initial_weights)
    num_samples = len(features)

    iteration = 0
    diff = float('inf')  # Set an initial large value to ensure the loop runs at least once

    # loop proceeds until either maximum number of iterations is reaches or the difference
    # becomes less than the desired error"""
    while iteration < num_iterations and diff > epsilon:

        gradient = compute_gradient(features=features, outcome=outcome,
                                    weights=weights, num_samples=num_samples)

        # Update weights using the gradient and learning rate
        weights -= learning_rate * gradient

        # if desired, I expedite the algorithm by setting all the weights except the first one to the
        # values obtained by the closed-form solution in each iteration
        if expedite_algorithm:
            weights[1:] = compute_weight_min_mse(data=data, poly_order=poly_order)[1:]

        # update iteration number and check if difference in weights is less than epsilon
        iteration += 1
        diff = np.sum(np.square(learning_rate * gradient))

        if return_log:
            all_weights = np.vstack(tup=(all_weights, weights))

    if return_log:
        # create dataframe
        col_names = ["{}{}".format('w', str(weight_num)) for weight_num in range(1, len(all_weights) + 1)]
        df_all_weights = pd.DataFrame(data=np.array(all_weights),
                                      columns=col_names)

        # insert iteration number
        df_all_weights.insert(0, "iteration_num", df_all_weights.index)

        return df_all_weights
    else:
        return weights



# read version from installed package
from importlib.metadata import version

__version__ = version("smltheory")

#__all__ = ['bayes_decision', 'est_error', 'generate_data', 'gradient_descent',
#           'least_squares', 'opt_error', 'overfitting']
#
#from smltheory import *
